@@ -6,7 +6,7 @@
 #ifndef _IP_VS_H
 #define _IP_VS_H
 
-#define IP_VS_VERSION_CODE            0x000801
+#define IP_VS_VERSION_CODE            0x000802
 #define NVERSION(version)                       \
         (version >> 16) & 0xFF,                 \
         (version >> 8) & 0xFF,                  \
@@ -111,25 +111,28 @@ struct ip_vs_rule_user {
 #include <asm/atomic.h>                 /* for struct atomic_t */
 #include <linux/netdevice.h>		/* for struct neighbour; */
 #include <net/dst.h>			/* for struct dst_entry */
+#include <linux/in6.h>
+#include <asm/uaccess.h>
+#include <asm/checksum.h>		/* for csum_partial */
 
 
 /*  #define CONFIG_IP_VS_DEBUG */
 
 #ifdef CONFIG_IP_VS_DEBUG
 extern int ip_vs_get_debug_level(void);
-#define IP_VS_DBG(level, msg...)                        \
-    do {                                                \
-            if (level <= ip_vs_get_debug_level())       \
-                    printk(KERN_DEBUG "IPVS: " ## msg); \
+#define IP_VS_DBG(level, msg...)			\
+    do {						\
+            if (level <= ip_vs_get_debug_level())	\
+                    printk(KERN_DEBUG "IPVS: " msg);	\
     } while (0)
 #else	/* NO DEBUGGING at ALL */
 #define IP_VS_DBG(level, msg...)  do {} while (0)
 #endif
 
-#define IP_VS_ERR(msg...) printk(KERN_ERR "IPVS: " ## msg )
-#define IP_VS_INFO(msg...) printk(KERN_INFO "IPVS: " ## msg )
+#define IP_VS_ERR(msg...) printk(KERN_ERR "IPVS: " msg)
+#define IP_VS_INFO(msg...) printk(KERN_INFO "IPVS: " msg)
 #define IP_VS_WARNING(msg...) \
-	printk(KERN_WARNING "IPVS: " ## msg)
+	printk(KERN_WARNING "IPVS: " msg)
 
 #ifdef CONFIG_IP_VS_DEBUG
 #define EnterFunction(level)                                            \
@@ -523,8 +526,6 @@ extern void ip_vs_app_cleanup(void);
  */
 extern int register_ip_vs_scheduler(struct ip_vs_scheduler *scheduler);
 extern int unregister_ip_vs_scheduler(struct ip_vs_scheduler *scheduler);
-extern int ip_vs_register_scheduler_module(struct ip_vs_scheduler *scheduler);
-extern void ip_vs_unregister_scheduler_module(struct ip_vs_scheduler *scheduler);
 extern int ip_vs_bind_scheduler(struct ip_vs_service *svc,
                                 struct ip_vs_scheduler *scheduler);
 extern int ip_vs_unbind_scheduler(struct ip_vs_service *svc);
@@ -605,6 +606,17 @@ extern __inline__ char ip_vs_fwd_tag(struct ip_vs_conn *cp)
 }
 
 
+/*
+ *	csum_partial wrapper, because csum_partial/i386 cannot provide
+ *	correct result when len=0 and buff isn't double word aligned.
+ */
+extern inline unsigned int
+ip_vs_csum_partial(const unsigned char * buff, int len, unsigned int sum)
+{
+	return len ? csum_partial(buff, len, sum) : sum;
+}
+
+
 /*
  *      Destination cache
  */
@@ -645,6 +657,29 @@ __ip_vs_dst_check(struct ip_vs_dest *dest, u32 rtos, u32 cookie)
         return dst;
 }
 
+/*
+ *	Own skb_cow() beast, tweaked for rewriting commonly
+ *	used pointers in the code
+ */
+static inline struct sk_buff *vs_skb_cow(struct sk_buff **skb_p,
+                                  struct iphdr **iph_p, unsigned char **t_p)
+{
+	struct sk_buff *skb=(*skb_p);
+	if (skb_cloned(skb)) {
+		skb = skb_copy(skb, GFP_ATOMIC);
+		if (skb) {
+			/*
+			 *	skb changed, update other pointers
+			 */
+			struct iphdr *iph = skb->nh.iph;
+			kfree_skb(*skb_p);
+			*skb_p = skb;
+			*iph_p = iph;
+			*t_p = (char*) iph + iph->ihl * 4;
+		}
+	}
+	return skb;
+}
 
 #endif /* __KERNEL__ */
 
