@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_conn.c,v 1.15.2.5 2001/06/22 15:03:08 wensong Exp $
+ * Version:     $Id: ip_vs_conn.c,v 1.15.2.6 2001/10/22 13:54:20 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -724,8 +724,7 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
         if ((iph->protocol != IPPROTO_UDP) || (h.uh->check != 0)) {
                 switch (skb->ip_summed) {
                 case CHECKSUM_NONE:
-                        csum = csum_partial(h.raw + doff, size - doff, 0);
-                        csum_ok++;
+			csum = ip_vs_csum_partial(h.raw+doff, size-doff, 0);
                         skb->csum = csum_partial(h.raw , doff, csum);
 
                 case CHECKSUM_HW:
@@ -787,6 +786,10 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
         dst_release(skb->dst);
 	skb->dst = &rt->u.dst;
 
+	/* copy-on-write the packet before mangling it */
+	if ((skb = vs_skb_cow(&skb, &iph, &h.raw)) == NULL)
+		return NF_DROP;
+
         /* mangle the packet */
         iph->daddr = cp->daddr;
         h.portp[1] = cp->dport;
@@ -816,9 +819,8 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
          */
         if (!csum_ok) {
                 /* Transport's payload partial checksum */
-                csum = csum_partial(h.raw + doff, size - doff, 0);
+                csum = ip_vs_csum_partial(h.raw + doff, size - doff, 0);
         }
-        skb->csum = csum;
 
         /*
          *	Protocol csum
@@ -826,15 +828,17 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
         switch (iph->protocol) {
         case IPPROTO_TCP:
                 h.th->check = 0;
+		skb->csum = csum_partial(h.raw, doff, csum);
                 h.th->check=csum_tcpudp_magic(iph->saddr, iph->daddr,
                                               size, iph->protocol,
-                                              csum_partial(h.raw, doff, csum));
+                                              skb->csum);
                 break;
         case IPPROTO_UDP:
                 h.uh->check = 0;
+		skb->csum = csum_partial(h.raw, doff, csum);
                 h.uh->check=csum_tcpudp_magic(iph->saddr, iph->daddr,
                                               size, iph->protocol,
-                                              csum_partial(h.raw, doff, csum));
+                                              skb->csum);
                 if (h.uh->check == 0)
                         h.uh->check = 0xFFFF;
                 break;
