@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_core.c,v 1.17.2.2 2001/06/22 15:03:08 wensong Exp $
+ * Version:     $Id: ip_vs_core.c,v 1.17.2.3 2001/10/22 13:54:20 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -93,7 +93,7 @@ int ip_vs_proto_doff(unsigned proto, char *th, unsigned size)
                  */
                 if (size >= sizeof(struct tcphdr)) {
                         ret = ((struct tcphdr*)th)->doff << 2;
-                        if (ret > size) {
+                        if (ret > size || ret < sizeof(struct tcphdr)) {
                                 ret = -1 ;
                         }
                 }
@@ -196,31 +196,6 @@ ip_vs_conn_stats(struct ip_vs_conn *cp, struct ip_vs_service *svc)
 }
 
 
-/*
- *	Own skb_cow() beast, tweaked for rewriting commonly
- *	used pointers in the code
- */
-static struct sk_buff *vs_skb_cow(struct sk_buff **skb_p,
-                                  struct iphdr **iph_p, unsigned char **t_p)
-{
-	struct sk_buff *skb=(*skb_p);
-	if (skb_cloned(skb)) {
-		skb = skb_copy(skb, GFP_ATOMIC);
-		if (skb) {
-			/*
-			 *	skb changed, update other pointers
-			 */
-			struct iphdr *iph = skb->nh.iph;
-			kfree_skb(*skb_p);
-			*skb_p = skb;
-			*iph_p = iph;
-			*t_p = (char*) iph + iph->ihl * 4;
-		}
-	}
-	return skb;
-}
-
-
 /*
  *  IPVS persistent scheduling function
  *  It creates a connection entry according to its template if exists,
@@ -709,6 +684,14 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	if (skb->nfcache & NFC_IPVS_PROPERTY)
                 return NF_ACCEPT;
 
+	if (skb->nh.iph->frag_off & htons(IP_MF|IP_OFFSET)) {
+		skb = ip_defrag(skb);
+		if (!skb) {
+			return NF_STOLEN;
+		}
+		*skb_p = skb;
+	}
+
 	iph = skb->nh.iph;
 	h.raw = (char*) iph + iph->ihl * 4;
 	size = ntohs(iph->tot_len) - (iph->ihl * 4);
@@ -766,10 +749,10 @@ static unsigned int ip_vs_out(unsigned int hooknum,
                 switch (skb->ip_summed)
                 {
                 case CHECKSUM_NONE:
-                        csum = csum_partial(h.raw + doff, size - doff, 0);
+			csum = ip_vs_csum_partial(h.raw+doff, size-doff, 0);
                         IP_VS_DBG(11, "O-pkt: %s I-datacsum=%d\n",
                                   ip_vs_proto_name(iph->protocol), csum);
-                        skb->csum = csum_partial(h.raw , doff, csum);
+                        skb->csum = csum_partial(h.raw, doff, csum);
                         csum_ok++;
                 case CHECKSUM_HW:
                         if (csum_tcpudp_magic(iph->saddr, iph->daddr, size,
@@ -829,9 +812,8 @@ static unsigned int ip_vs_out(unsigned int hooknum,
  	 */
 	if (!csum_ok) {
                 /* transport's payload partial csum */
-		csum = csum_partial(h.raw + doff, size - doff, 0);
+		csum = ip_vs_csum_partial(h.raw + doff, size - doff, 0);
 	}
-	skb->csum = csum;
 
 	IP_VS_DBG(11, "O-pkt: %s size=%d O-datacsum=%d\n",
                   ip_vs_proto_name(iph->protocol), size, csum);
@@ -842,18 +824,20 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	switch (iph->protocol) {
         case IPPROTO_TCP:
                 h.th->check = 0;
+		skb->csum = csum_partial(h.raw, doff, csum);
                 h.th->check=csum_tcpudp_magic(iph->saddr, iph->daddr,
                                               size, iph->protocol,
-                                              csum_partial(h.raw, doff, csum));
+                                              skb->csum);
                 IP_VS_DBG(11, "O-pkt: %s O-csum=%d (+%d)\n",
                           ip_vs_proto_name(iph->protocol), h.th->check,
                           (char*) & (h.th->check) - (char*) h.raw);
                 break;
         case IPPROTO_UDP:
                 h.uh->check = 0;
+		skb->csum = csum_partial(h.raw, doff, csum);
                 h.uh->check=csum_tcpudp_magic(iph->saddr, iph->daddr,
                                               size, iph->protocol,
-                                              csum_partial(h.raw, doff, csum));
+                                              skb->csum);
                 if (h.uh->check == 0)
                         h.uh->check = 0xFFFF;
                 IP_VS_DBG(11, "O-pkt: %s O-csum=%d (+%d)\n",
